@@ -2,8 +2,8 @@
 
 # Copyright (C) 2002-2015 Jörg Platte <joergplatte@gmx.de>
 # Copyright (C) 2002-2015 Cajus Pollmeier <pollmeier@gonicus.de>
-# Copyright (C) 2013-2015 Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
 # Copyright (C) 2013-2015 Alex Owen <r.alex.owen@gmail.com>
+# Copyright (C) 2013-2022 Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,9 +18,6 @@
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
-#
-# On Debian GNU/Linux systems, a copy of the GNU General Public License may be
-# found in the file /usr/share/common-licenses/GPL.
 
 use strict;
 my $LDAPENABLED = eval "use Net::LDAP; 1" ? '1' : '0';
@@ -36,20 +33,29 @@ my $configfile="/etc/uif/uif.conf";
 my $configfile6="/etc/uif/uif6.conf";
 my $ipv6=0;
 
+my @icmpv6_types_mapping = ( [ 'neighbor-advertisement',  'nd-neighbor-advert'  ],
+                             [ 'neighbour-advertisement', 'nd-neighbor-advert'  ],
+                             [ 'neighbor-solicitation',   'nd-neighbor-solicit' ],
+                             [ 'neighbour-solicitation',  'nd-neighbor-solicit' ],
+                             [ 'router-advertisement',    'nd-router-advert' ],
+                             [ 'router-solicitation',     'nd-router-solicit' ]);
+
 my @mapping = (	[ 'n', 'uifid', 'Name' ],
 		[ 's', 'uifsource', 'Source'],
 		[ 'i', 'uifindevice', 'InputInterface'],
-		[ 'pi', 'uifpindevice', 'PhysicalInputInterface'],
 		[ 'd', 'uifdest', 'Destination'],
 		[ 'o', 'uifoutdevice', 'OutputInterface'],
-		[ 'po', 'uifpoutdevice', 'PhysicalOutputInterface'],
 		[ 'p', 'uifservice', 'Service'],
 		[ 'm', 'uifmark', 'MarkMatch'],
 		[ 'S', 'uiftranssource', 'TranslatedSource'],
 		[ 'D', 'uiftransdest', 'TranslatedDestination'],
 		[ 'P', 'uiftransservice', 'TranslatedService'],
 		[ '',  'uiftype', 'Type'],
-		[ 'f', 'uifflag', 'Flags']);
+		[ 'f', 'uifflag', 'Flags'],
+# only supported with old iptables backend... deprecated with nftables
+# FIXME: disable these options in the uif.conf parser code at some point
+		[ 'pi', 'uifpindevice', 'PhysicalInputInterface'],
+		[ 'po', 'uifpoutdevice', 'PhysicalOutputInterface']);
 
 my %charstringmap;
 my %ldapstringmap;
@@ -72,7 +78,6 @@ sub readConfig {
 	my $state='NONE';
 	my $line;
 
-
 	unless (defined($$Protocols{'OK'})) {
 		$$Protocols{'OK'}=1;
 		open (PROT, '/etc/protocols') || die "Can't read '/etc/protocols'\n";
@@ -118,16 +123,24 @@ sub readConfig {
 				}
 			} elsif ($state eq 'INCLUDE') {
 				if ($line =~ /^\s*\"(.+)\"$/) {
-					my $file = $1;
-					readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+					my $filename_pattern = $1;
+					my @filenames = glob($filename_pattern);
+					my $file;
+					foreach $file(@filenames) {
+						readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+					}
 				} else {
 					die "invalid line in section include: $line\n";
 				}
 			} elsif ($state eq 'INCLUDE6') {
 				if ($ipv6) {
 					if ($line =~ /^\s*\"(.+)\"$/) {
-						my $file = $1;
-						readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+						my $filename_pattern = $1;
+						my @filenames = glob($filename_pattern);
+						my $file;
+						foreach $file(@filenames) {
+							readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+						}
 					} else {
 						die "invalid line in section include6: $line\n";
 					}
@@ -135,8 +148,12 @@ sub readConfig {
 			} elsif ($state eq 'INCLUDE4') {
 				if ($ipv6) {} else {
 					if ($line =~ /^\s*\"(.+)\"$/) {
-						my $file = $1;
-						readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+						my $filename_pattern = $1;
+						my @filenames = glob($filename_pattern);
+						my $file;
+						foreach $file(@filenames) {
+							readConfig ($file, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Id, $Sysconfig);
+						}
 					} else {
 						die "invalid line in section include4: $line\n";
 					}
@@ -456,13 +473,27 @@ sub validateSysconfig {
 			} else {
 				die "invalid account prefix: $prefix\n";
 			}
+		} elsif ("\L$syskey" eq "filtercommand") {
+			my $filter_command=$$Sysconfig{$syskey};
+			delete $$Sysconfig{$syskey};
+			$filter_command =~ s/\s+//g;
+			if ($filter_command eq 'nft') {
+				$$Sysconfig{'FilterCommand'} = 'nft';
+			} elsif ($filter_command eq 'iptables') {
+				$$Sysconfig{'FilterCommand'} = 'iptables';
+			} elsif ($filter_command eq 'iptables-nft') {
+				$$Sysconfig{'FilterCommand'} = 'iptables-nft';
+			} elsif ($filter_command eq 'iptables-legacy') {
+				$$Sysconfig{'FilterCommand'} = 'iptables-legacy';
+			} else {
+				die "invalid packet filter command, use 'nft', 'iptables', 'iptables-nft' or 'iptables-legacy'";
+			}
 		} else {
 			die "unknown sysconfig parameter: $syskey\n";
 		}
 	}
 }
 
-
 sub toRange {
 	my ($range, $proto, $rule) = @_;
 
@@ -621,11 +652,11 @@ sub validateData {
 						my $only_proto = $1;
 						$position =~ s/\((.+)\)$//;
 						if (($ipv6) && ($only_proto eq "4")) {
-							print "IPv6 setup: Skipping IPv4-only rule for network \"$position\"\n";
+							print STDERR "IPv6 setup: Skipping IPv4-only rule for network \"$position\"\n";
 							$$rule{'Type'} = 'IGNORE-IPV4-ONLY';
 							next;
 						} elsif ((! $ipv6) && ($only_proto eq "6")) {
-							print "IPv4 setup: Skipping IPv6-only rule for network \"$position\"\n";
+							print STDERR "IPv4 setup: Skipping IPv6-only rule for network \"$position\"\n";
 							$$rule{'Type'} = 'IGNORE-IPV6-ONLY';
 							next;
 						}
@@ -1019,7 +1050,464 @@ sub validateData {
 	}
 }
 
-sub genRuleDump {
+sub icmpv6_types_IPTABLES_to_NFT {
+	my ($type) = @_;
+	my $nft_type = $type;
+	foreach my $map (@icmpv6_types_mapping) {
+		if ( "@$map[0]" eq "$type" ) {
+			$nft_type = "@$map[1]";
+			last;
+		}
+	}
+	return $nft_type;
+}
+
+sub genRuleDump_NFT {
+	my ($Rules, $Listing, $Sysconfig) = @_;
+	my @partial;
+	my $rule;
+	my %nat;
+	my %filter;
+	my %mangle;
+	my @nat;
+	my @filter;
+	my @mangle;
+	my $table;
+	my $chains;
+	my $inet;
+
+	if ($ipv6) {
+		$inet = "ip6";
+	} else {
+		$inet = "ip";
+	}
+
+	foreach $rule (@$Rules) {
+
+		if ( ($$rule{'Type'} eq "IGNORE-IPV4-ONLY") || ($$rule{'Type'} eq "IGNORE-IPV6-ONLY") ) {
+			next;
+		}
+
+		my @protocol;
+		my @source;
+		my @destination;
+		my @inputinterface;
+		my @outputinterface;
+		my @mark;
+		my $action;
+		my $logaction;
+		my $type;
+		my $name;
+		my $proto;
+		my $id;
+		my $not;
+
+		if ($$rule{'Table'} eq 'filter') {
+			$table=\@filter;
+			$chains=\%filter;
+		} elsif ($$rule{'Table'} eq 'nat') {
+			$table=\@nat;
+			$chains=\%nat;
+		} elsif ($$rule{'Table'} eq 'mangle') {
+			$table=\@mangle;
+			$chains=\%mangle;
+		} else {
+			die "$$rule{'Table'} is not implemented!\n";
+		}
+
+		$type="add rule $inet $$rule{'Table'} $$rule{'Type'}";
+		if (exists($$rule{'Name'})) {
+			$name=$$rule{'Name'};
+			$name=~s/\s+//g;
+		} else {
+			$name="-";
+		}
+		$id=$$rule{'Id'};
+
+		if (exists($$rule{'Reject'})) {
+			if ($$rule{'Reject'} ne '1') {
+				$action="counter reject with $$rule{'Reject'}";
+				$action =~ s/-/ /g;
+			} else {
+				$action="counter jump MYREJECT";
+			}
+			$logaction="REJECT";
+		} elsif ($$rule{'Action'} eq "TCPMSS") {
+			$action="tcp flags & (syn|rst) == syn tcp option maxseg size >= 1400 counter tcp option maxseg size set rt mtu";
+			$logaction="TCPMSS";
+		} elsif ($$rule{'Action'} eq "MARK") {
+			$action="counter meta mark set $$rule{'Mark'}";
+			$logaction="MARK";
+		} else {
+			if ( $$rule{'Action'} =~ /^(ACCEPT|DROP|RETURN|MASQUERADE|DNAT|SNAT)$/) {
+				$action="counter ".lc $$rule{'Action'};
+			} else {
+				$action="counter jump $$rule{'Action'}";
+			}
+			$logaction=$$rule{'Action'};
+		}
+
+		if (exists($$rule{"Service-not"})) {
+			$not='!';
+		} else {
+			$not='';
+		}
+
+		foreach $proto (qw(tcp udp)) {
+			if (exists($$rule{"\u$proto"})) {
+				my $string;
+				my $entry;
+				foreach $entry (qw(0 1)) {
+					my $multiport;
+					my $count=0;
+					foreach $multiport (@{$$rule{"\u$proto"}[$entry]}) {
+						if ($count==0) {
+							$string='';
+						}
+						$string.="$multiport,";
+						$count++;
+						if ($count==15) {
+							$string =~ s/,$//;
+							$string="meta l4proto $proto ".($entry==1?"d":"s")."port {".$string."}";
+							push (@protocol, $string);
+							$string='';
+							$count=0;
+						}
+					}
+					if (defined($string) && $count) {
+						$string =~ s/,$//;
+						if ($count > 1) {
+							$string="meta l4proto $proto $proto ".($entry==1?"d":"s")."port {".$string."}";
+						} else {
+							$string="meta l4proto $proto $proto ".($entry==1?"d":"s")."port {".$string."}";
+						}
+						push (@protocol, $string);
+					}
+				}
+				my $range;
+				foreach $range (@{$$rule{"\u$proto"}[2]}) {
+					$range =~ s/\:/-/g;
+					$range =~ /^(.+)-(.+)$/;
+					if ( $1 eq $2 ) {
+						$range = $1;
+					}
+					push (@protocol, "$not $proto sport $range");
+				}
+				foreach $range (@{$$rule{"\u$proto"}[3]}) {
+					$range =~ s/\:/-/g;
+					$range =~ /^(.+)-(.+)$/;
+					if ( $1 eq $2 ) {
+						$range = $1;
+					}
+					push (@protocol, "$not $proto dport $range");
+				}
+				my $range_sport;
+				my $range_dport;
+				foreach $range (@{$$rule{"\u$proto"}[4]}) {
+					$range =~ s/\:/-/g;
+					$range =~ /^(.+)\/(.+)$/;
+					$range_sport = $1;
+					$range_dport = $2;
+					$range_sport =~ /^(.+)-(.+)$/;
+					if ( $1 eq $2 ) {
+						$range_sport = $1;
+					}
+					$range_dport =~ /^(.+)-(.+)$/;
+					if ( $1 eq $2 ) {
+						$range_dport = $1;
+					}
+					push (@protocol, "$not $proto sport $range_sport $not $proto dport $range_dport");
+				}
+			}
+		}
+		if (exists($$rule{'ICMP'}) && (! $ipv6)) {
+			my $type;
+			foreach $type (@{$$rule{'ICMP'}}) {
+				if ($type eq 'all') {
+					push (@protocol, "$not icmp");
+				} else {
+					push (@protocol, "icmp $not type $type");
+				}
+			}
+		}
+		if (exists($$rule{'ICMP6'}) and ($ipv6)) {
+			my $type;
+			foreach $type (@{$$rule{'ICMP6'}}) {
+				if ($type eq 'all') {
+					push (@protocol, "meta l4proto $not ipv6-icmp");
+				} else {
+					my $nft_type = icmpv6_types_IPTABLES_to_NFT($type);
+					push (@protocol, "icmpv6 type $not $nft_type");
+				}
+			}
+		}
+		if (exists($$rule{'OtherProtocols'})) {
+			my $proto;
+			foreach $proto (@{$$rule{'OtherProtocols'}}) {
+				push (@protocol, "$not $proto");
+			}
+		}
+		if (exists($$rule{'Source'})) {
+			if (exists($$rule{'Source-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $source;
+			foreach $source (@{$$rule{'Source'}}) {
+				if ($source =~ /(.+)=(.+)/ && ($$rule{'Table'} eq 'filter')) {
+					push (@source, "$not $inet saddr $1 mac $not ether saddr $2 counter");
+				} else {
+					$source =~ /([^=]+)/;
+					push (@source, "$not $inet saddr $1");
+				}
+			}
+		}
+		if (exists($$rule{'Destination'})) {
+			if (exists($$rule{'Destination-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $destination;
+			foreach $destination (@{$$rule{'Destination'}}) {
+				$destination =~ /([^=]+)/;
+				push (@destination, "$not $inet daddr $1");
+			}
+		}
+		if (exists($$rule{'TranslatedSource'})) {
+			my $source;
+			$source=${$$rule{'TranslatedSource'}}[0];
+			$source =~ /([^=]+)/;
+			$source=$1;
+			my $ip = NetAddr::IP->new($source) || die "not a valid network: $source\n";
+			my $net=$ip->network();
+			my $bcast = $ip->broadcast();
+			if ($net ne $bcast) {
+				$source="$net-$bcast";
+			}
+			$source =~ s/\/[^-]+//g;
+			$action.=" to $source";
+		}
+		if (exists($$rule{'TranslatedDestination'})) {
+			my $destination;
+			$destination=${$$rule{'TranslatedDestination'}}[0];
+			$destination =~ /([^=]+)/;
+			$destination=$1;
+			my $ip = NetAddr::IP->new($destination) || die "not a valid network: $destination\n";
+			my $net=$ip->network();
+			my $bcast = $ip->broadcast();
+			if ($net ne $bcast) {
+				$destination="$net-$bcast";
+			}
+			$destination =~ s/\/[^-]+//g;
+			$action.=" to $destination";
+		}
+
+		foreach $proto (qw(tcp udp)) {
+			if (exists($$rule{"Translated\u$proto"})) {
+				my $ref = $$rule{"Translated\u$proto"};
+				if (defined($$ref[1][0])) {
+					$action.=":$$ref[1][0]";
+				}
+				if (defined($$ref[3][0])) {
+					$action.=":$$ref[3][0]";
+				}
+				last;
+			}
+		}
+
+		if (exists($$rule{'InputInterface'})) {
+			if (exists($$rule{'InputInterface-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $input;
+			foreach $input (@{$$rule{'InputInterface'}}) {
+				push (@inputinterface, "$not iifname \"$input\"");
+			}
+		}
+		if (exists($$rule{'OutputInterface'})) {
+			if (exists($$rule{'OutputInterface-not'})) {
+				$not='!';
+			} else {
+				$not='';
+			}
+			my $output;
+			foreach $output (@{$$rule{'OutputInterface'}}) {
+				push (@outputinterface, "$not oifname \"$output\"");
+			}
+		}
+		if (exists($$rule{'MarkMatch'})) {
+			my $mark;
+			foreach $mark (@{$$rule{'MarkMatch'}}) {
+				push (@mark, "meta mark set $mark");
+
+			}
+		}
+
+		if (exists($$rule{'Log'})) {
+			my $chain = "${id}$$rule{'Action'}log";
+			$$chains{$chain}=1;
+			my $logid;
+			if ($$rule{'Log'}) {
+				$logid=$$rule{'Log'};
+			} else {
+				$logid=$name;
+			}
+			push (@$table, "add rule $inet $$rule{'Table'} CHAIN_$chain limit rate $$Sysconfig{'LogLimit'} burst $$Sysconfig{'LogBurst'} packets counter log prefix \"$$Sysconfig{'LogPrefix'} $logaction ($logid): \" level $$Sysconfig{'LogLevel'} flags tcp options flags ip options");
+			push (@$table, "add rule $inet $$rule{'Table'} CHAIN_$chain $action");
+			$action="counter jump CHAIN_$chain";
+		}
+		if (exists($$rule{'Accounting'})) {
+			my $accountchain="$$Sysconfig{'AccountPrefix'}$$rule{'Accounting'}";
+			unless (exists($$chains{"$accountchain"})) {
+				$$chains{"$accountchain"}=1;
+				push (@$table, "add rule $inet $$rule{'Table'} CHAIN_$accountchain $action");
+			}
+			my $accountrules="${id}_ACCOUNTING_$$rule{'Accounting'}";
+			$$chains{$accountrules}=1;
+			push (@$table, "$type counter jump $accountrules");
+			push (@$table, "add rule $inet $$rule{'Table'} ACCOUNTING$$rule{'Type'} counter jump CHAIN_$accountrules");
+			$type="add rule $inet $$rule{'Table'} $accountrules ";
+			$action=" counter jump CHAIN_$accountchain";
+		}
+		if (exists($$rule{'Limit'})) {
+			$action=" limit rate $$rule{'Limit'} burst $$rule{'Limit-burst'} packets $action";
+		}
+		my @rulearray = (\@inputinterface, \@outputinterface, \@protocol, \@source, \@destination, \@mark);
+
+		my $level=1;
+		my $again=1;
+		while ($again) {
+			@partial = ();
+			$again=0;
+			my $array;
+			# adjust if you have many entries...
+			my $depth=0xFFFF;
+			foreach $array (@rulearray) {
+				if (@$array && $depth>@$array) {
+					$depth=@$array;
+				}
+			}
+			foreach $array (@rulearray) {
+				if (@$array==$depth) {
+					my $i;
+					for ($i=0; $i<@$array; $i++) {
+						$partial[$i].=" $$array[$i]";
+					}
+					@$array = ();
+					if ($depth != 1) {
+						last;
+					}
+				}
+			}
+			foreach $array (@rulearray) {
+				if (@$array) {
+					$again=1;
+					last;
+				}
+			}
+			my $jumpto;
+			if ($again) {
+				$jumpto="counter jump CHAIN_${id}_$level";
+			} else {
+				$jumpto=$action;
+			}
+			if (@partial) {
+				my $newjumpto;
+				my $part;
+				foreach $part (@partial) {
+					$newjumpto=$jumpto;
+					if ($part =~ /-p (udp|tcp)/ && $jumpto =~ /-p (udp|tcp)/) {
+						$newjumpto =~ s/-p (udp|tcp) -m (udp|tcp)//;
+					}
+					push (@$table, $type." $part $newjumpto");
+				}
+			} else {
+				push (@$table, "$type $jumpto");
+			}
+			if ($again) {
+				$type="add rule $inet $$rule{'Table'} CHAIN_${id}_$level";
+				$$chains{"${id}_$level"}=1;
+				$level++;
+			}
+		}
+	}
+
+	# make sure, all rules get dropped before populating tables anew
+	push (@$Listing, "flush ruleset $inet");
+
+	my $entry;
+	foreach $entry (qw(mangle filter nat)) {
+		if ($entry eq "nat" && $ipv6 == 1) {next};
+		my $chain;
+		push (@$Listing, "add table $inet $entry");
+		if ($entry eq 'filter') {
+			$table=\@filter;
+			$chains=\%filter;
+			push (@$Listing, "add chain $inet filter MYREJECT");
+			push (@$Listing, "add chain $inet filter STATENOTNEW");
+			foreach (qw(INPUT OUTPUT FORWARD)) {
+				push (@$Listing, "add chain $inet filter ACCOUNTING$_");
+				push (@$Listing, "add chain $inet filter ACCOUNTINGSTATELESS$_");
+				push (@$Listing, "add chain $inet filter STATE$_");
+				push (@$Listing, "add chain $inet filter STATELESS$_");
+				push (@$Listing, "add chain $inet filter $_ { type filter hook ".lc $_." priority 0; policy drop; }");
+				push (@$Listing, "add rule $inet filter $_ counter jump STATE$_");
+				push (@$Listing, "add rule $inet filter STATE$_ ct state invalid counter jump STATELESS$_");
+				push (@$Listing, "add rule $inet filter STATE$_ counter jump ACCOUNTING$_");
+				push (@$Listing, "add rule $inet filter STATE$_ ct state related,established counter accept");
+				if ($ipv6) {
+					push (@$Listing, "add rule ip6 filter STATE$_ meta l4proto != ipv6-icmp ct state != new counter jump STATENOTNEW");
+				} else {
+					push (@$Listing, "add rule ip filter STATE$_ ct state != new counter jump STATENOTNEW");
+				}
+				push (@$Listing, "add rule $inet filter STATELESS$_ counter jump ACCOUNTINGSTATELESS$_");
+			}
+			push (@$Listing, "add rule $inet filter STATENOTNEW limit rate $$Sysconfig{'LogLimit'} burst $$Sysconfig{'LogBurst'} packets counter log prefix \"$$Sysconfig{'LogPrefix'} STATE NOT NEW: \" level $$Sysconfig{'LogLevel'} flags tcp options flags ip options");
+			push (@$Listing, "add rule $inet filter STATENOTNEW counter drop");
+			push (@$Listing, "add rule $inet filter MYREJECT counter reject with tcp reset");
+			push (@$Listing, "add rule $inet filter MYREJECT counter reject");
+		} elsif ($entry eq 'nat') {
+			$table=\@nat;
+			$chains=\%nat;
+			push (@$Listing, "add chain $inet nat POSTROUTING { type nat hook postrouting priority srcnat; policy accept; }");
+			foreach (qw(PREROUTING)) {
+				push (@$Listing, "add chain $inet nat $_ { type nat hook ".lc $_." priority dstnat; policy accept; }");
+			}
+			foreach (qw(OUTPUT)) {
+				push (@$Listing, "add chain $inet nat $_ { type nat hook ".lc $_." priority -100; policy accept; }");
+			}
+		} else {
+			$table=\@mangle;
+			$chains=\%mangle;
+			push (@$Listing, "add chain $inet mangle PREROUTING { type filter hook prerouting priority mangle; policy accept; }");
+			push (@$Listing, "add chain $inet mangle OUTPUT { type route hook output priority mangle; policy accept; }");
+		}
+		foreach (keys(%$chains)) {
+			push (@$Listing, "add chain $inet filter CHAIN_$_");
+		}
+		push (@$Listing, "#");
+		push (@$Listing, "# beginning of user generated $entry rules");
+		push (@$Listing, "#");
+		foreach (@$table) {
+			push (@$Listing, $_);
+		}
+		push (@$Listing, "#");
+		push (@$Listing, "# end of user generated $entry rules");
+		push (@$Listing, "#");
+		if ($entry eq 'filter') {
+			foreach (qw(INPUT OUTPUT FORWARD)) {
+				push (@$Listing, "add rule $inet filter STATELESS$_ limit rate $$Sysconfig{'LogLimit'} burst $$Sysconfig{'LogBurst'} packets counter log prefix \"$$Sysconfig{'LogPrefix'} INVALID STATE: \"  level $$Sysconfig{'LogLevel'} flags tcp options flags ip options");
+				push (@$Listing, "add rule $inet filter STATELESS$_ counter drop");
+			}
+		}
+	}
+}
+
+sub genRuleDump_IPTABLES {
 	my ($Rules, $Listing, $Sysconfig) = @_;
 	my @partial;
 	my $rule;
@@ -1309,22 +1797,22 @@ sub genRuleDump {
 			} else {
 				$logid=$name;
 			}
-			push (@$table, "-A $chain -m limit --limit $$Sysconfig{'LogLimit'} --limit-burst $$Sysconfig{'LogBurst'} -j LOG --log-prefix \"$$Sysconfig{'LogPrefix'} $logaction ($logid): \" --log-level $$Sysconfig{'LogLevel'} --log-tcp-options --log-ip-options");
-			push (@$table, "-A $chain $action");
-			$action="-j $chain";
+			push (@$table, "-A CHAIN_$chain -m limit --limit $$Sysconfig{'LogLimit'} --limit-burst $$Sysconfig{'LogBurst'} -j LOG --log-prefix \"$$Sysconfig{'LogPrefix'} $logaction ($logid): \" --log-level $$Sysconfig{'LogLevel'} --log-tcp-options --log-ip-options");
+			push (@$table, "-A CHAIN_$chain $action");
+			$action="-j CHAIN_$chain";
 		}
 		if (exists($$rule{'Accounting'})) {
 			my $accountchain="$$Sysconfig{'AccountPrefix'}$$rule{'Accounting'}";
 			unless (exists($$chains{"$accountchain"})) {
 				$$chains{"$accountchain"}=1;
-				push (@$table, "-A $accountchain $action");
+				push (@$table, "-A CHAIN_$accountchain $action");
 			}
 			my $accountrules="${id}_ACCOUNTING_$$rule{'Accounting'}";
 			$$chains{$accountrules}=1;
 			push (@$table, "$type -j $accountrules");
-			push (@$table, "-A ACCOUNTING$$rule{'Type'} -j $accountrules");
+			push (@$table, "-A ACCOUNTING$$rule{'Type'} -j CHAIN_$accountrules");
 			$type="-A $accountrules ";
-			$action=" -j $accountchain";
+			$action=" -j CHAIN_$accountchain";
 		}
 		if (exists($$rule{'Limit'})) {
 			$action=" -m limit --limit $$rule{'Limit'} --limit-burst $$rule{'Limit-burst'} $action";
@@ -1364,7 +1852,7 @@ sub genRuleDump {
 			}
 			my $jumpto;
 			if ($again) {
-				$jumpto="-j ${id}_$level";
+				$jumpto="-j CHAIN_${id}_$level";
 			} else {
 				$jumpto=$action;
 			}
@@ -1382,7 +1870,7 @@ sub genRuleDump {
 				push (@$table, "$type $jumpto");
 			}
 			if ($again) {
-				$type="-A ${id}_$level";
+				$type="-A CHAIN_${id}_$level";
 				$$chains{"${id}_$level"}=1;
 				$level++;
 			}
@@ -1438,7 +1926,7 @@ sub genRuleDump {
 			}
 		}
 		foreach (keys(%$chains)) {
-			push (@$Listing, ":$_ - [0:0]");
+			push (@$Listing, ":CHAIN_$_ - [0:0]");
 		}
 		push (@$Listing, "#");
 		push (@$Listing, "# beginning of user generated $entry rules");
@@ -1465,20 +1953,93 @@ sub printRules {
 	print @$Listing;
 }
 
+sub printRulesWithLinenumbers {
+	my ($Listing) = @_;
+	@$Listing=map { $_."\n" } @$Listing;
+	my $i = 1;
+	for (@$Listing) {
+		print $i++,":","\t",$_ if $_ ne '';
+	}
+}
+
 sub signalCatcher {
 	$SignalCatched=1;
 }
-sub applyRules {
-	my ($timeout, $Listing) = @_;
+
+sub applyRules_NFT {
+	my ($timeout, $Listing, $Sysconfig) = @_;
 	my @oldrules;
 	my $error;
 
 	@$Listing=map { $_."\n" } @$Listing;
-	if ($ipv6) {
-		open (IPT, '/sbin/ip6tables-save|');
-	} else {
-		open (IPT, '/sbin/iptables-save|');
+	open (NFT, '/usr/sbin/nft list ruleset|');
+	@oldrules = <NFT>;
+	close (NFT);
+
+	$SIG{'INT'} = 'signalCatcher';
+	$SIG{'KILL'} = 'signalCatcher';
+	$SIG{'QUIT'} = 'signalCatcher';
+	$SIG{'TERM'} = 'signalCatcher';
+
+	open (NFT_DISK, '>/tmp/nftrules');
+	print NFT_DISK @$Listing;
+	close (NFT_DISK);
+
+	open (NFT, '|/usr/sbin/nft -f -');
+	print NFT @$Listing;
+	close (NFT);
+	$error=$?;
+
+	if ($timeout && !$error) {
+		sleep $timeout;
 	}
+	if ($timeout || $SignalCatched || $error) {
+		open (NFT, '|/usr/sbin/nft -f -');
+		print NFT @oldrules;
+		close (NFT);
+		if ($SignalCatched) {
+			die "aborted. old rules restored.\n";
+		} elsif ($error) {
+			die "error in generated rules\n";
+		}
+	}
+}
+
+sub applyRules_IPTABLES {
+	my ($timeout, $Listing, $Sysconfig) = @_;
+	my @oldrules;
+	my $error;
+	my $save_cmd;
+	my $restore_cmd;
+
+	if (%$Sysconfig{'FilterCommand'} eq 'iptables') {
+		if ($ipv6) {
+			$save_cmd    = "/usr/sbin/ip6tables-save";
+			$restore_cmd = "/usr/sbin/ip6tables-restore";
+		} else {
+			$save_cmd    = "/usr/sbin/iptables-save";
+			$restore_cmd = "/usr/sbin/iptables-restore";
+		}
+	} elsif (%$Sysconfig{'FilterCommand'} eq 'iptables-nft') {
+		if ($ipv6) {
+			$save_cmd    = "/usr/sbin/ip6tables-nft-save";
+			$restore_cmd = "/usr/sbin/ip6tables-nft-restore";
+		} else {
+			$save_cmd    = "/usr/sbin/iptables-nft-save";
+			$restore_cmd = "/usr/sbin/iptables-nft-restore";
+		}
+	} elsif (%$Sysconfig{'FilterCommand'} eq 'iptables-legacy') {
+		if ($ipv6) {
+			$save_cmd    = "/usr/sbin/ip6tables-legacy-save";
+			$restore_cmd = "/usr/sbin/ip6tables-legacy-restore";
+		} else {
+			$save_cmd    = "/usr/sbin/iptables-legacy-save";
+			$restore_cmd = "/usr/sbin/iptables-legacy-restore";
+		}
+	}
+
+	@$Listing=map { $_."\n" } @$Listing;
+	open (IPT, "$save_cmd|");
 	@oldrules = <IPT>;
 	close (IPT);
 
@@ -1487,11 +2048,7 @@ sub applyRules {
 	$SIG{'QUIT'} = 'signalCatcher';
 	$SIG{'TERM'} = 'signalCatcher';
 
-	if ($ipv6) {
-		open (IPT, '|/sbin/ip6tables-restore');
-	} else {
-		open (IPT, '|/sbin/iptables-restore');
-	}
+	open (IPT, "|$restore_cmd");
 	print IPT @$Listing;
 	close (IPT);
 	$error=$?;
@@ -1500,11 +2057,7 @@ sub applyRules {
 		sleep $timeout;
 	}
 	if ($timeout || $SignalCatched || $error) {
-		if ($ipv6) {
-			open (IPT, '|/sbin/ip6tables-restore');
-		} else {
-			open (IPT, '|/sbin/iptables-restore');
-		}
+		open (IPT, "|$restore_cmd");
 		print IPT @oldrules;
 		close (IPT);
 		if ($SignalCatched) {
@@ -1526,6 +2079,7 @@ sub readCommandLine {
 	my @Listing;
 	my $test=0;
 	my $print=0;
+	my $linenumbers=0;
 	my $disable=0;
 	my $writeconfigfile;
 	my $writeldapruleset;
@@ -1540,6 +2094,11 @@ sub readCommandLine {
 	my $ldappassword;
 	my $timeout=0;
 
+	if (exists($ENV{'FILTER_COMMAND'})) {
+		$Sysconfig{'FilterCommand'}=$ENV{'FILTER_COMMAND'};
+	} else {
+		$Sysconfig{'FilterCommand'}='nft';
+	}
 	if (exists($ENV{'LOGLIMIT'})) {
 		$Sysconfig{'LogLimit'}=$ENV{'LOGLIMIT'};
 	} else {
@@ -1576,13 +2135,14 @@ sub readCommandLine {
 		$Sysconfig{'AccountPrefix'}='ACC_';
 	}
 	my %opt;
-	getopts('6c:tpds:b:r:T:C:R:D:w:W', \%opt) || uifUsg ();
+	getopts('6c:tplds:b:r:T:C:R:D:w:W', \%opt) || uifUsg ();
 
 	$ipv6 = 1 if $opt{'6'};
 	$configfile=$configfile6 if $opt{'6'};
 	$configfile = $opt{'c'} if $opt{'c'};
 	$test = 1 if $opt{'t'};
 	$print = 1 if $opt{'p'};
+	$linenumbers = 1 if $opt{'l'};
 	$disable = 1 if $opt{'d'};
 	if ($opt{'T'}) {
 		if ($opt{'T'} =~ /^(\d+)$/) {
@@ -1662,21 +2222,46 @@ sub readCommandLine {
 			exit 0;
 		} else {
 			validateData (\%Networks, \%Services, \%Interfaces, \%Protocols, \@Rules, \%Sysconfig, \%Marker);
-			genRuleDump (\@Rules, \@Listing, \%Sysconfig);
+			if ($Sysconfig{'FilterCommand'} eq 'nft') {
+				genRuleDump_NFT (\@Rules, \@Listing, \%Sysconfig);
+			} else {
+				genRuleDump_IPTABLES (\@Rules, \@Listing, \%Sysconfig);
+			}
 		}
 	} else {
-		clearAllRules (\@Listing);
+		validateSysconfig (\%Sysconfig);
+		if ($Sysconfig{'FilterCommand'} eq 'nft') {
+			clearAllRules_NFT (\@Listing);
+		} else {
+			clearAllRules_IPTABLES (\@Listing);
+		}
 	}
 
-	if ($print) {
+	if ($print && $linenumbers) {
+		printRulesWithLinenumbers (\@Listing);
+	} elsif ($print) {
 		printRules (\@Listing);
 	}
 	if ($test==0) {
-		applyRules ($timeout, \@Listing);
+		if ($Sysconfig{'FilterCommand'} eq 'nft') {
+			applyRules_NFT ($timeout, \@Listing, \%Sysconfig);
+		} else {
+			applyRules_IPTABLES ($timeout, \@Listing, \%Sysconfig);
+		}
 	}
 }
 
-sub clearAllRules {
+sub clearAllRules_NFT {
+	my ($Listing) = @_;
+
+	if ($ipv6) {
+		push (@$Listing,"flush ruleset ip6");
+	} else {
+		push (@$Listing,"flush ruleset ip");
+	}
+}
+
+sub clearAllRules_IPTABLES {
 	my ($Listing) = @_;
 
 	push (@$Listing,"*mangle");
@@ -1703,6 +2288,7 @@ sub uifUsg {
 	print "-c  read <configfile> instead of $configfile (or in ipv6 mode $configfile6)\n";
 	print "-t  test rules\n";
 	print "-p  print rules to stdout\n";
+	print "-l  if '-p' is used, prepend line numbers to the rules printout\n";
 	print "-d  disable firewall (clear all rules)\n";
 	print "-s  LDAP-server (default: localhost)\n";
 	print "-b  LDAP-base\n";
@@ -1800,7 +2386,6 @@ sub readLdap {
 		}
 	}
 
-
 	$mesg = $ldap->search (	base => $ldapbase,
 				filter => "objectClass=ipNetwork");
 	if ($mesg->is_error) {
@@ -2102,7 +2687,6 @@ sub addLdap {
 	}
 }
 
-
 sub writeConfig {
 	my ($configfile, $Networks, $Services, $Interfaces, $Protocols, $Rules, $Sysconfig, $Marker) = @_;
 
